// **************************************************************************
//
//
// **************************************************************************
//  Written by HairyDairyMaid (a.k.a. - lightbulb)
//  hairydairymaid@yahoo.com
// **************************************************************************
//
//  This program is copyright (C) 2004 HairyDairyMaid (a.k.a. Lightbulb)
//  This program is free software; you can redistribute it and/or modify it
//  under the terms of version 2 the GNU General Public License as published
//  by the Free Software Foundation.
//  This program is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
//  more details.
//  To view a copy of the license go to:
//  http://www.fsf.org/copyleft/gpl.html
//  To receive a copy of the GNU General Public License write the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// **************************************************************************

// Default is Compile for Linux (both #define's below should be commented out)
#define WINDOWS_VERSION   // uncomment only this for Windows Compile / MS Visual C Compiler
// #define __FreeBSD__       // uncomment only this for FreeBSD

#ifdef WINDOWS_VERSION
   #include <windows.h>      // Only for Windows Compile
   #define strcasecmp  stricmp
   #define strncasecmp strnicmp
#endif

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "x300tpatch.h"

static unsigned int ctrl_reg;

int pfd;
int instruction_length;
int issue_reset      = 1;
int issue_enable_mw  = 1;
int issue_watchdog   = 1;
int issue_break      = 1;
int issue_erase      = 1;
int issue_timestamp  = 1;
int force_dma        = 0;
int force_nodma      = 0;
int selected_fc      = 0;
unsigned int selected_window  = 0;
unsigned int selected_start   = 0;
unsigned int selected_length  = 0;
int custom_options   = 0;
int silent_mode      = 0;
int skipdetect       = 0;
int instrlen         = 0;
int wiggler          = 0;


char            flash_part[128];
unsigned int    flash_size = 0;

int             block_total = 0;
unsigned int    block_addr = 0;
unsigned int    blocks[1024];
unsigned int    cmd_type = 0;

char            AREA_NAME[128];
unsigned int    AREA_START;
unsigned int    AREA_LENGTH;
unsigned int    FLASH_MEMORY_START;
unsigned int    vendid;
unsigned int    devid;

unsigned int    data_register;
unsigned int    address_register;

int USE_DMA       = 0;
int ejtag_version = 0;


typedef struct _processor_chip_type {
    unsigned int        chip_id;        // Processor Chip ID
    int                 instr_length;   // EJTAG Instruction Length
    char*               chip_descr;     // Processor Chip Description
} processor_chip_type;


processor_chip_type  processor_chip_list[] = {
   { 0x08630001, 5, "SigmaDesigns SMP8634 Rev A CPU" },
   { 0, 0, 0 }
   };


// -----------------------------------------
// ---- Start of Compiler Specific Code ----
// -----------------------------------------


void lpt_openport(void)
{
   #ifdef WINDOWS_VERSION    // ---- Compiler Specific Code ----

      HANDLE h;

      h = CreateFile("\\\\.\\giveio", GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
      if(h == INVALID_HANDLE_VALUE) {  printf("Couldn't access giveio device\n");   CloseHandle(h);   exit(0);   }
      CloseHandle(h);

   #else                     // ---- Compiler Specific Code ----

      #ifdef __FreeBSD__     // ---- Compiler Specific Code ----

         pfd = open("/dev/ppi0", O_RDWR);
         if (pfd < 0)   {   perror("Failed to open /dev/ppi0");   exit(0);   }
         if ((ioctl(pfd, PPEXCL) < 0) || (ioctl(pfd, PPCLAIM) < 0))  {   perror("Failed to lock /dev/ppi0");   close(pfd);   exit(0);   }

      #else                  // ---- Compiler Specific Code ----

         pfd = open("/dev/parport0", O_RDWR);
         if (pfd < 0)   {   perror("Failed to open /dev/parport0");   exit(0);   }
         if ((ioctl(pfd, PPEXCL) < 0) || (ioctl(pfd, PPCLAIM) < 0))   {   perror("Failed to lock /dev/parport0");   close(pfd);   exit(0);   }

      #endif

   #endif
}


void lpt_closeport(void)
{
   #ifndef WINDOWS_VERSION   // ---- Compiler Specific Code ----

      #ifndef __FreeBSD__    // ---- Compiler Specific Code ----

         if (ioctl(pfd, PPRELEASE) < 0)  {  perror("Failed to release /dev/parport0");  close(pfd);  exit(0);  }

      #endif

      close(pfd);

   #endif
}


static unsigned char clockin(int tms, int tdi)
{
   unsigned char data;

   tms = tms ? 1 : 0;
   tdi = tdi ? 1 : 0;
   	
   if(wiggler) data = (1 << WTDO) | (0 << WTCK) | (tms << WTMS) | (tdi << WTDI) | (1 << WTRST_N);
   else        data = (1 << TDO) | (0 << TCK) | (tms << TMS) | (tdi << TDI);
   #ifdef WINDOWS_VERSION   // ---- Compiler Specific Code ----  
      _outp(0x378, data);  
   #else  
      ioctl(pfd, PPWDATA, &data);  
   #endif

   if(wiggler) data = (1 << WTDO) | (1 << WTCK) | (tms << WTMS) | (tdi << WTDI) | (1 << WTRST_N);
   else        data = (1 << TDO) | (1 << TCK) | (tms << TMS) | (tdi << TDI);
   #ifdef WINDOWS_VERSION   // ---- Compiler Specific Code ----  
      _outp(0x378, data);  
   #else  
      ioctl(pfd, PPWDATA, &data);  
   #endif

   #ifdef WINDOWS_VERSION   // ---- Compiler Specific Code ----  
      data = (unsigned char)_inp(0x379);  
   #else  
      ioctl(pfd, PPRSTATUS, &data);  
   #endif

   data ^= 0x80;
   data >>= (wiggler ? WTDO : TDO);
   data &= 1;

   return data;
}


// ---------------------------------------
// ---- End of Compiler Specific Code ----
// ---------------------------------------


void test_reset(void)
{
    clockin(1, 0);  // Run through a handful of clock cycles with TMS high to make sure
    clockin(1, 0);  // we are in the TEST-LOGIC-RESET state.
    clockin(1, 0);
    clockin(1, 0);
    clockin(1, 0);
    clockin(0, 0);  // enter runtest-idle
}


void set_instr(int instr)
{
    int i;
    static int curinstr = 0xFFFFFFFF;

    if (instr == curinstr)
       return;

    clockin(1, 0);  // enter select-dr-scan
    clockin(1, 0);  // enter select-ir-scan
    clockin(0, 0);  // enter capture-ir
    clockin(0, 0);  // enter shift-ir (dummy)
    for (i=0; i < instruction_length; i++)
    {
        clockin(i==(instruction_length - 1), (instr>>i)&1);
    }
    clockin(1, 0);  // enter update-ir
    clockin(0, 0);  // enter runtest-idle

    curinstr = instr;
}


static unsigned int ReadWriteData(unsigned int in_data)
{
    int i;
    unsigned int out_data = 0;
    unsigned char out_bit;

    clockin(1, 0);  // enter select-dr-scan
    clockin(0, 0);  // enter capture-dr
    clockin(0, 0);  // enter shift-dr
    for(i = 0 ; i < 32 ; i++)
    {
       out_bit  = clockin((i == 31), ((in_data >> i) & 1));
       out_data = out_data | (out_bit << i);
    }
    clockin(1,0);   // enter update-dr
    clockin(0,0);   // enter runtest-idle
    return out_data;
}


static unsigned int ReadData(void)
{
    return ReadWriteData(0x00);
}


void WriteData(unsigned int in_data)
{
    ReadWriteData(in_data);
}


void ShowData(unsigned int value)
{
    int i;
    for (i=0; i<32; i++)
        printf("%d", (value >> (31-i)) & 1);
    printf(" (%08X)\n", value);
}


static unsigned int ejtag_read(unsigned int addr)
{
   return(ejtag_pracc_read(addr));
}


static unsigned int ejtag_read_h(unsigned int addr)
{
   return(ejtag_pracc_read_h(addr));
}


void ejtag_write(unsigned int addr, unsigned int data)
{
   ejtag_pracc_write(addr, data);
}


void ejtag_write_h(unsigned int addr, unsigned int data)
{
   ejtag_pracc_write_h(addr, data);
}


static unsigned int ejtag_pracc_read(unsigned int addr)
{
   address_register = addr | 0xA0000000;  // Force to use uncached segment
   data_register    = 0x0;
   ExecuteDebugModule(pracc_readword_code_module);
   return(data_register);
}


void ejtag_pracc_write(unsigned int addr, unsigned int data)
{
   address_register = addr | 0xA0000000;  // Force to use uncached segment
   data_register    = data;
   ExecuteDebugModule(pracc_writeword_code_module);
}


static unsigned int ejtag_pracc_read_h(unsigned int addr)
{
   address_register = addr | 0xA0000000;  // Force to use uncached segment
   data_register    = 0x0;
   ExecuteDebugModule(pracc_readhalf_code_module);
   return(data_register);
}


void ejtag_pracc_write_h(unsigned int addr, unsigned int data)
{
   address_register = addr | 0xA0000000;  // Force to use uncached segment
   data_register    = data;
   ExecuteDebugModule(pracc_writehalf_code_module);
}


void ExecuteDebugModule(unsigned int *pmodule)
{
   unsigned int ctrl_reg;
   unsigned int address;
   unsigned int data   = 0;
   unsigned int offset = 0;
   int finished = 0;
   int DEBUGMSG = 0;
      
   if (DEBUGMSG) printf("DEBUGMODULE: Start module.\n");
   
   // Feed the chip an array of 32 bit values into the processor via the EJTAG port as instructions.
   while (1)
   {
      // Read the control register.  Make sure an access is requested, then do it.
      while(1) 
      {
         set_instr(INSTR_CONTROL);
         ctrl_reg = ReadWriteData(PRACC | PROBEN | SETDEV);
         if (ctrl_reg & PRACC)
            break;
         if (DEBUGMSG) printf("DEBUGMODULE: No memory access in progress!\n");
      }
      
      set_instr(INSTR_ADDRESS);
      address = ReadData();
      
      // Check for read or write
      if (ctrl_reg & PRNW) // Bit set for a WRITE
      {
         // Read the data out
         set_instr(INSTR_DATA);
         data = ReadData();
      
         // Clear the access pending bit (let the processor eat!)
         set_instr(INSTR_CONTROL);
         ctrl_reg = ReadWriteData(PROBEN | SETDEV);
      
         // Processor is writing to us
         if (DEBUGMSG) printf("DEBUGMODULE: Write 0x%08X to address 0x%08X\n", data, address);
         // Handle Debug Write
         // If processor is writing to one of our psuedo virtual registers then save off data
         if (address == MIPS_VIRTUAL_ADDRESS_ACCESS)  address_register = data;
         if (address == MIPS_VIRTUAL_DATA_ACCESS)     data_register    = data;
      }
      
      else
      
      {
         // Check to see if its reading at the debug vector.  The first pass through
         // the module is always read at the vector, so the first one we allow.  When
         // the second read from the vector occurs we are done and just exit.
         if (address == MIPS_DEBUG_VECTOR_ADDRESS)
         {
            if (finished++) // Allows ONE pass
            {
               if (DEBUGMSG) printf("DEBUGMODULE: Finished module.\n");
               return;
            }
         }
      
         // Processor is reading from us
         if (address >= MIPS_DEBUG_VECTOR_ADDRESS)
         {
            // Reading an instruction from our module so fetch the instruction from the module
            offset = (address - MIPS_DEBUG_VECTOR_ADDRESS) / 4;
            data = *(unsigned int *)(pmodule + offset);
            if (DEBUGMSG) printf("DEBUGMODULE: Instruction read at 0x%08X  offset -> %04d  data -> 0x%08X\n", address, offset, data); //fflush(stdout);
         }
         else
         {
            // Reading from our virtual register area
            if (DEBUGMSG) printf("DEBUGMODULE: Read address 0x%08X  data = 0x%08X\n", address, data);
            // Handle Debug Read
            // If processor is reading from one of our psuedo virtual registers then give it data
            if (address == MIPS_VIRTUAL_ADDRESS_ACCESS)  data = address_register;
            if (address == MIPS_VIRTUAL_DATA_ACCESS)     data = data_register;
         }
      
         // Send the data out
         set_instr(INSTR_DATA);
         data = ReadWriteData(data);
      
         // Clear the access pending bit (let the processor eat!)
         set_instr(INSTR_CONTROL);
         ctrl_reg = ReadWriteData(PROBEN | SETDEV);
      
      }
   }
}


void chip_detect(void)
{
    unsigned int id = 0x0;
    
    processor_chip_type*   processor_chip = processor_chip_list;

    lpt_openport();

    printf("Probing bus ... ");
    
	 // Auto Detect CPU Chip ID
	while (processor_chip->chip_id)
	{
	  test_reset();
	  if (instrlen) 
		 instruction_length = instrlen;
	  else 
		 instruction_length = processor_chip->instr_length;
	  set_instr(INSTR_IDCODE);
	  id = ReadData();
	  if (id == processor_chip->chip_id)
	  {
		 printf("Done\n\n");
		 printf("CPU Chip ID: ");  ShowData(id);  printf("*** Found a %s chip ***\n\n", processor_chip->chip_descr);
		 return;
	  }
	  processor_chip++;
	}

    printf("Done\n\n");
    printf("CPU Chip ID: ");  ShowData(id);  printf("*** Unknown or NO CPU Chip ID Detected ***\n\n");

	if (id == 0xffffffff)
	{
	    printf("*** Possible Causes:\n");
	    printf("    1) X300T is not Connected.\n");
	    printf("    2) X300T is not Powered On.\n");
	    printf("    3) Improper JTAG Cable.\n");

	    chip_shutdown();;
	    exit(0);
	}
}


void check_ejtag_features() 
{
    unsigned int features;
    
    set_instr(INSTR_IMPCODE);
    features = ReadData();
}


unsigned int exitDebugMode (void)
{
  unsigned int ui_result;

  // Feed the processor a DERET instruction.

  // Read the control register and make sure processor is attempting to 
  // read from dmseg.
  ui_result = dMIPS32_ECR_CPU_RESET_OCCURED | dMIPS32_ECR_ACCESS_PENDING |
              dMIPS32_ECR_PROBE_ENABLE | dMIPS32_ECR_PROBE_VECTOR;

  set_instr(dJTAG_REGISTER_CONTROL);
  ui_result = ReadWriteData(ui_result);

  if (ui_result & dMIPS32_ECR_ACCESS_PENDING)
  {
    // Make sure processor is not writing
    if (ui_result & dMIPS32_ECR_ACCESS_RW) // Bit set for a WRITE
    {
      // Chip is writing, need to clear the write
      printf("EJTAGEMUL ERROR: Processor writing on debug exit.\n");  
      return(0);
    }

    // OK - Processor is reading, make sure its in the instruction space
    set_instr(dJTAG_REGISTER_ADDRESS);
    ui_result = ReadWriteData(0x00); // Address is 0-write only
    if (ui_result > dMIPS32_DMSEG_RAM_BEGIN)
    {
      // Chip is not reading from the instruction area!
      printf("JTAGEMUL ERROR: Processor not reading instruction for DERET!\n");
      return(0);
    }

    // Chip is reading from instruction space - feed it a DERET
    // Write the DERET instruction into the data register            
    set_instr(dJTAG_REGISTER_DATA);
    ReadWriteData(dMIPS32_DERET_INSTRUCTION);

    // Now finish the processor access by clearing the access pending bit
    set_instr(dJTAG_REGISTER_CONTROL);

    ui_result = dMIPS32_ECR_CPU_RESET_OCCURED | dMIPS32_ECR_PROBE_ENABLE | 
                dMIPS32_ECR_PROBE_VECTOR;

    ui_result = ReadWriteData(ui_result);

    // At this point we are done.  Even if the chip didn't exit debug mode
    // the debug mode catcher will grab it and report the reason to GDB.
    printf("JTAGEMUL: Processor exited debug mode.\n");
    return(1);
  }
  else
  {
    // ERROR!  What to do if the processor is not accessing?
    printf("EJTAGEMUL ERROR: Processor not accessing for debug exit.\n");
    return(0);
  }

  // Should never get here!  
  return(0); 
}

void chip_shutdown(void)
{
	exitDebugMode();
	
    fflush(stdout);
    test_reset();
    lpt_closeport();
}


void apply_patch(char *filename, unsigned int start, unsigned int length)
{
    unsigned int addr, data1, data2, data3, data4;
    FILE *fd;
    int counter = 0;
    int percent_complete = 0;
	int speed = 0;
    char newfilename[128] = "";
    time_t start_time = time(0);
	time_t elapsed_time;
    time_t end_time, elapsed_seconds;

    struct tm* lt = localtime(&start_time);
    char time_str[15];
    
    printf("=========================\n");
    printf("Patching Started\n");
    printf("=========================\n");

	addr = 0x9363fe88;
    data1 = ejtag_read(addr);
    data2 = ejtag_read(addr+4);
    printf("0x9363fe88: %x%x\n", data1, data2);
	if (data1 == 0x0cd8dc32 && data2 == 0x24040001)
	{
		printf("patch1 location verified\n");
		ejtag_write(addr, 0x08480000);  //  jal 0x936370C8   -->  j 0x91200000   (yamon entry point)
		ejtag_write(addr+4, 0x00);
	}
	else
	{
		printf("patch1 location not found. wrong bootloader version ?\n");
	}
	
	
	addr = 0x9363ff88;
    data3 = ejtag_read(addr);
    data4 = ejtag_read(addr+4);
    printf("0x9363ff88: %x%x\n", data3, data4);
	if (data3 == 0x0cd8dc32 && data4 == 0x24040001)  
	{
		printf("patch2 location verified\n");
		ejtag_write(addr, 0x08480000);   //  jal 0x936370C8   -->  j 0x91200000   (yamon entry point)
		ejtag_write(addr+4, 0x00);
	}
	else
	{
		printf("patch2 location not found. wrong bootloader version ?\n");
	}
	
	
	// signature skip flag
	addr = 0x937E1728;
	ejtag_write(addr, 0x01000000);
	
    printf("\nDone  \n\n");
}

void show_usage(void)
{
   processor_chip_type*  processor_chip = processor_chip_list;
   int counter = 0;
   printf( " USAGE: x300yamon  \n"
           "            /wiggler ........... use wiggler cable\n\n"
           "            -----------------------------------------------\n");
}

void loadYamon(void)
{
	FILE *fd;
    int counter = 0;
    int percent_complete = 0;
	unsigned int start, data, addr, length;
	
    fd=fopen("yamon.bin", "rb" );
    if (fd<=0)
    {
        fprintf(stderr,"Could not open yamon.bin for reading\n");
        exit(1);
    }

    printf("=========================\n");
    printf("uploading yamon to RAM\n");
    printf("=========================\n");

	start = 0x91200000;
	length = 453056;

    for(addr=start; addr<(start+length); addr+=4)
    {
        counter += 4;
        percent_complete = (counter * 100 / length);
        if ((addr&0xF) == 0)  printf("[%3d%% uploaded]   %08x \r", percent_complete, addr);

        fread( (unsigned char*) &data, 1,sizeof(data), fd);
		ejtag_write(addr, data);
        
        fflush(stdout);         	
        data = 0xFFFFFFFF;  // This is in case file is shorter than expected length
    }
    fclose(fd);
}

int main(int argc, char** argv)
{
    char choice[128];
    int run_option;
    int j;

    printf("\n");
    printf("===============================================\n");
    printf("X300t yamon loader v1.0  by MCE2222\n");
	printf("based on WRT54G/GS EJTAG Debrick Utility which \n");
	printf("is actually derived from Mips32 EJTAG from \n");
	printf("TotalEmbedded without giving credit... ts ts ts\n");
    printf("===============================================\n");
    printf("= This version only works on bootloader V1051 =\n");
    printf("===============================================\n\n");

	show_usage();
    
    run_option = 1;
	selected_start = 0x93600000;
	selected_length = 0xf0000;

    if (argc > 1)
    {
       j = 1;
       while (j < argc)
       {
          strcpy(choice,argv[j]);

          if (strcasecmp(choice,"/wiggler")==0)         
		  {
				printf("\nusing Wiggler interface\n");
				wiggler = 1;
		  }
          else
          {
             show_usage();
             printf("\n*** ERROR - Invalid <option> specified ***\n\n");
             exit(1);
          }
          j++;
       }
    }
	
    instrlen = 5;
    silent_mode = 1;
	skipdetect = 1;	   
    
    // ----------------------------------
    // Detect CPU 
    // ----------------------------------
    chip_detect();


    // ----------------------------------
    // Find Implemented EJTAG Features
    // ----------------------------------
    check_ejtag_features();


    // ----------------------------------
    // Reset State Machine For Good Measure
    // ----------------------------------
    test_reset();


    // ----------------------------------
    // Reset Processor and Peripherals
    // ----------------------------------
    printf("Issuing Processor / Peripheral Reset ... ");
    if (issue_reset)
    {
       set_instr(INSTR_CONTROL);
       ctrl_reg = ReadWriteData(PRRST | PERRST);
       printf("Done\n");
    } else printf("Skipped\n");


    // ----------------------------------
    // Put into EJTAG Debug Mode
    // ----------------------------------
    printf("Halting Processor ... ");
    if (issue_break)
    {
       set_instr(INSTR_CONTROL);
       ctrl_reg = ReadWriteData(PRACC | PROBEN | SETDEV | JTAGBRK );
       if (ReadWriteData(PRACC | PROBEN | SETDEV) & BRKST)  
          printf("<Processor Entered Debug Mode!> ... ");
       else  
          printf("<Processor did NOT enter Debug Mode!> ... ");
       printf("Done\n");
    } else printf("Skipped\n");


    // ----------------------------------
    // Clear Watchdog
    // ----------------------------------
    printf("Clearing Watchdog ... ");        
    if (issue_watchdog)
    {
       ejtag_write(0xb8000080,0);
       printf("Done\n");
    } else printf("Skipped\n");



    // ----------------------------------
    // Execute Requested Operation
    // ----------------------------------
	loadYamon();
    apply_patch(AREA_NAME, selected_start, selected_length);

    printf("\n\n *** REQUESTED OPERATION IS COMPLETE ***\n\n");

    chip_shutdown();

    return 0;
}


// **************************************************************************
// End of File
// **************************************************************************
